#include "demo.h"

#define RELOAD_CANVAS() UnloadTexture(occlusionMap.tex); \
                        occlusionMap.tex = LoadTextureFromImage(occlusionMap.img); \
                        UnloadTexture(emissionMap.tex); \
                        emissionMap.tex = LoadTextureFromImage(emissionMap.img);

Demo::Demo() {
  // --- MISC PARAMETERS
  maxRaySteps = 64;
  jfaSteps = 128;

  orbs = false;
  srgb = false;

  gi = false;
  giRayCount = 64;
  giNoise = true;
  giMixFactor = 0.7;
  giDecayRate = 1.3;

  rcRayCount = 4;
  cascadeAmount = 5;
  cascadeDisplayIndex = 0;
  rcBilinear = true;

  user.mode = DRAWING;
  userSetRandomColor();

  // UI
  skipUIRendering = false;
  debugShowBuffers = false;
  debug = false;
  help  = true;

  debugWindowData.flags |= ImGuiWindowFlags_NoTitleBar;
  debugWindowData.flags |= ImGuiWindowFlags_NoScrollbar;
  debugWindowData.flags |= ImGuiWindowFlags_NoMove;
  debugWindowData.flags |= ImGuiWindowFlags_NoResize;
  debugWindowData.flags |= ImGuiWindowFlags_NoCollapse;
  debugWindowData.flags |= ImGuiWindowFlags_NoBackground;
  debugWindowData.flags |= ImGuiWindowFlags_NoNav;

  ImGuiIO& io = ImGui::GetIO();
  io.IniFilename = NULL;

  HideCursor();

  // --- LOAD RESOURCES

  lastMousePos = {0, 0};

  cursorTex = LoadTexture("res/textures/cursor.png");

  user.brush.img = LoadImage("res/textures/brush.png");
  user.brush.tex = LoadTextureFromImage(user.brush.img);
  user.brushSize = 0.25;

  setScene(CLEAR);

  // automatically load fragment shaders in the `res/shaders` directory
  FilePathList shaderFiles = LoadDirectoryFilesEx("res/shaders", ".frag", false);
  for (int i = 0; i < shaderFiles.count; i++) {
    std::string str = shaderFiles.paths[i];
    str.erase(0, 12);
    if (str != "broken.frag") loadShader(str);
  }
  UnloadDirectoryFiles(shaderFiles);

  // change bit depth for bufferA, B, & C so that we can encode texture coordinates without losing data
  // default Raylib FBOs have a bit depth of 8 per channel, which would only cover for a window of maximum size 255x255
  // we can also save some memory by reducing bit depth of buffers to what is strictly required
  auto changeBitDepth = [](RenderTexture2D &buffer, PixelFormat pixformat) {
    rlEnableFramebuffer(buffer.id);
      rlUnloadTexture(buffer.texture.id);
      buffer.texture.id = rlLoadTexture(NULL, SCREEN_WIDTH, SCREEN_HEIGHT, pixformat, 1);
      buffer.texture.width = SCREEN_WIDTH;
      buffer.texture.height = SCREEN_HEIGHT;
      buffer.texture.format = pixformat;
      buffer.texture.mipmaps = 1;
      rlFramebufferAttach(buffer.id, buffer.texture.id, RL_ATTACHMENT_COLOR_CHANNEL0, RL_ATTACHMENT_TEXTURE2D, 0);
    rlDisableFramebuffer();
  };

  bufferA         = LoadRenderTexture(SCREEN_WIDTH, SCREEN_HEIGHT);
  bufferB         = LoadRenderTexture(SCREEN_WIDTH, SCREEN_HEIGHT);
  bufferC         = bufferA;
  radianceBufferA = LoadRenderTexture(SCREEN_WIDTH, SCREEN_HEIGHT);
  radianceBufferB = LoadRenderTexture(SCREEN_WIDTH, SCREEN_HEIGHT);
  radianceBufferC = radianceBufferA;
  sceneBuf        = LoadRenderTexture(SCREEN_WIDTH, SCREEN_HEIGHT);
  tempBuf        = LoadRenderTexture(SCREEN_WIDTH, SCREEN_HEIGHT);
  distFieldBuf    = LoadRenderTexture(SCREEN_WIDTH, SCREEN_HEIGHT);
  lastFrameBuf    = LoadRenderTexture(SCREEN_WIDTH, SCREEN_HEIGHT);

  changeBitDepth(bufferA,      PIXELFORMAT_UNCOMPRESSED_R32G32B32A32);
  changeBitDepth(bufferB,      PIXELFORMAT_UNCOMPRESSED_R32G32B32A32);
  changeBitDepth(bufferC,      PIXELFORMAT_UNCOMPRESSED_R32G32B32A32);
  changeBitDepth(sceneBuf,     PIXELFORMAT_UNCOMPRESSED_R5G5B5A1);
  changeBitDepth(distFieldBuf, PIXELFORMAT_UNCOMPRESSED_R16);
}

// -------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void Demo::update() {
}

// -------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void Demo::render() {
  ClearBackground(PINK);

  const Shader& rcShader        = shaders["rc.frag"];
  const Shader& giShader        = shaders["gi.frag"];
  const Shader& jfaShader       = shaders["jfa.frag"];
  const Shader& prepJfaShader   = shaders["prepjfa.frag"];
  const Shader& scenePrepShader = shaders["prepscene.frag"];
  const Shader& distFieldShader = shaders["distfield.frag"];
  const Shader& prepGiShader    = shaders["prepgi.frag"];
  const Shader& finalShader     = shaders["final.frag"];

  Vector2 resolution = { SCREEN_WIDTH, SCREEN_HEIGHT };

  // create scene texture - combining emission & occlusion maps into one texture
  // this is also the step to add dynamic gpu-driven lighting
  BeginTextureMode(sceneBuf);
    ClearBackground(BLANK);
    BeginShaderMode(scenePrepShader);
      float time = GetTime();
      Vector2 mousePos = GetMousePosition();
      int orbsInt = orbs;
      SetShaderValueTexture(scenePrepShader, GetShaderLocation(scenePrepShader, "uOcclusionMap"),    occlusionMap.tex);
      SetShaderValueTexture(scenePrepShader, GetShaderLocation(scenePrepShader, "uEmissionMap"),     emissionMap.tex);
      SetShaderValue(scenePrepShader, GetShaderLocation(scenePrepShader, "uMousePos"),   &mousePos, SHADER_UNIFORM_VEC2);
      SetShaderValue(scenePrepShader, GetShaderLocation(scenePrepShader, "uResolution"), &resolution, SHADER_UNIFORM_VEC2);
      SetShaderValue(scenePrepShader, GetShaderLocation(scenePrepShader, "uTime"),       &time,       SHADER_UNIFORM_FLOAT);
      SetShaderValue(scenePrepShader, GetShaderLocation(scenePrepShader, "uOrbs"),       &orbsInt,       SHADER_UNIFORM_INT);
      DrawRectangle(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, WHITE);
    EndShaderMode();
  EndTextureMode();

  // first render pass for JFA
  // create UV mask w/ prep shader
  BeginTextureMode(bufferA);
    ClearBackground(BLANK);
    BeginShaderMode(prepJfaShader);
      SetShaderValueTexture(prepJfaShader, GetShaderLocation(prepJfaShader, "uSceneMap"), sceneBuf.texture);
      SetShaderValue(prepJfaShader, GetShaderLocation(prepJfaShader, "uResolution"), &resolution, SHADER_UNIFORM_VEC2);
      DrawRectangle(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, WHITE);
    EndShaderMode();
  EndTextureMode();

  // // ping-pong buffering
  // // alternate between two buffers so that we can implement a recursive shader
  // // see https://mini.gmshaders.com/p/gm-shaders-mini-recursive-shaders-1308459
  for (int j = jfaSteps*2; j >= 1; j /= 2) {
    bufferC = bufferA;
    bufferA = bufferB;
    bufferB = bufferC;

    BeginTextureMode(bufferA);
      ClearBackground(BLANK);
      BeginShaderMode(jfaShader);
        SetShaderValueTexture(jfaShader, GetShaderLocation(jfaShader, "uCanvas"), bufferB.texture);
        SetShaderValue(jfaShader, GetShaderLocation(jfaShader, "uJumpSize"), &j, SHADER_UNIFORM_INT);
        SetShaderValue(jfaShader, GetShaderLocation(jfaShader, "uResolution"), &resolution, SHADER_UNIFORM_VEC2);
        DrawRectangle(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, WHITE);
      EndShaderMode();
    EndTextureMode();
  }

  // write distance field to another buffer
  // reduces strain as the cpu gets to send less data to the gpu for GI shader
  BeginTextureMode(distFieldBuf);
    ClearBackground(BLANK);
    BeginShaderMode(distFieldShader);
      SetShaderValueTexture(distFieldShader, GetShaderLocation(distFieldShader, "uJFA"), bufferA.texture);
      SetShaderValue(distFieldShader, GetShaderLocation(distFieldShader, "uResolution"), &resolution, SHADER_UNIFORM_VEC2);
      DrawRectangle(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, WHITE);
    EndShaderMode();
  EndTextureMode();

  int flipY = 1;
  int srgbInt = srgb;
  if (gi) {
    // mix last frame w/ current frame for indirect lighting
    BeginTextureMode(radianceBufferB);
      BeginShaderMode(prepGiShader);
        SetShaderValueTexture(prepGiShader, GetShaderLocation(prepGiShader, "uSceneMap"), sceneBuf.texture);
        SetShaderValueTexture(prepGiShader, GetShaderLocation(prepGiShader, "uLastFrame"), lastFrameBuf.texture);
        SetShaderValue(prepGiShader, GetShaderLocation(prepGiShader, "uResolution"), &resolution,  SHADER_UNIFORM_VEC2);
        SetShaderValue(prepGiShader, GetShaderLocation(prepGiShader, "uMixFactor"),  &giMixFactor, SHADER_UNIFORM_FLOAT);
        SetShaderValue(prepGiShader, GetShaderLocation(prepGiShader, "uFlipY"),  &flipY, SHADER_UNIFORM_INT);
        DrawRectangle(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, WHITE);
      EndShaderMode();
    EndTextureMode();

    int giNoiseInt = giNoise;
    BeginTextureMode(radianceBufferA);
      BeginShaderMode(giShader);
        ClearBackground(BLANK);
        SetShaderValueTexture(giShader, GetShaderLocation(giShader, "uDistanceField"), distFieldBuf.texture);
        SetShaderValueTexture(giShader, GetShaderLocation(giShader, "uSceneMap"),      sceneBuf.texture);
        SetShaderValueTexture(giShader, GetShaderLocation(giShader, "uLastFrame"),     radianceBufferB.texture);
        SetShaderValue(giShader, GetShaderLocation(giShader, "uResolution"), &resolution,  SHADER_UNIFORM_VEC2);
        SetShaderValue(giShader, GetShaderLocation(giShader, "uRaysPerPx"),  &giRayCount,  SHADER_UNIFORM_INT);
        SetShaderValue(giShader, GetShaderLocation(giShader, "uMaxSteps"),   &maxRaySteps, SHADER_UNIFORM_INT);
        SetShaderValue(giShader, GetShaderLocation(giShader, "uSrgb"),       &srgbInt,     SHADER_UNIFORM_INT);
        SetShaderValue(giShader, GetShaderLocation(giShader, "uNoise"),      &giNoiseInt,  SHADER_UNIFORM_INT);
        SetShaderValue(giShader, GetShaderLocation(giShader, "uDecayRate"),  &giDecayRate, SHADER_UNIFORM_FLOAT);
        DrawRectangle(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, WHITE);
      EndShaderMode();
    EndTextureMode();
  } else {

    int test = 0;

    for (int i = cascadeAmount; i >= 0; i--) {
      radianceBufferC = radianceBufferA;
      radianceBufferA = radianceBufferB;
      radianceBufferB = radianceBufferC;

      if (rcBilinear)
        SetTextureFilter(radianceBufferC.texture, TEXTURE_FILTER_BILINEAR);
      else
        SetTextureFilter(radianceBufferC.texture, TEXTURE_FILTER_POINT);

      BeginTextureMode(radianceBufferA);
        BeginShaderMode(rcShader);
          ClearBackground(BLANK);
          SetShaderValueTexture(rcShader, GetShaderLocation(rcShader, "uDistanceField"), distFieldBuf.texture);
          SetShaderValueTexture(rcShader, GetShaderLocation(rcShader, "uSceneMap"),      sceneBuf.texture);
          SetShaderValueTexture(rcShader, GetShaderLocation(rcShader, "uLastPass"),      radianceBufferC.texture);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uResolution"),          &resolution,          SHADER_UNIFORM_VEC2);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uBaseRayCount"),        &rcRayCount,          SHADER_UNIFORM_INT);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uMaxSteps"),            &maxRaySteps,         SHADER_UNIFORM_INT);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uCascadeDisplayIndex"), &cascadeDisplayIndex, SHADER_UNIFORM_INT);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uCascadeIndex"),        &i,                   SHADER_UNIFORM_INT);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uCascadeAmount"),       &cascadeAmount,       SHADER_UNIFORM_INT);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uSrgb"),                &srgbInt,             SHADER_UNIFORM_INT);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uTest"),                &test,                SHADER_UNIFORM_INT);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uDecayRate"),  &giDecayRate, SHADER_UNIFORM_FLOAT);
          DrawRectangle(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, WHITE);
        EndShaderMode();
      EndTextureMode();
    }

    flipY = 0;
    BeginTextureMode(tempBuf);
      // Vector2 res = {resolution.x, -resolution.y};
      BeginShaderMode(prepGiShader);
        SetShaderValueTexture(prepGiShader, GetShaderLocation(prepGiShader, "uSceneMap"), sceneBuf.texture);
        SetShaderValueTexture(prepGiShader, GetShaderLocation(prepGiShader, "uLastFrame"), radianceBufferA.texture);
        SetShaderValue(prepGiShader, GetShaderLocation(prepGiShader, "uResolution"), &resolution,  SHADER_UNIFORM_VEC2);
        SetShaderValue(prepGiShader, GetShaderLocation(prepGiShader, "uMixFactor"),  &giMixFactor, SHADER_UNIFORM_FLOAT);
        SetShaderValue(prepGiShader, GetShaderLocation(prepGiShader, "uFlipY"),  &flipY, SHADER_UNIFORM_INT);
        DrawRectangle(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, WHITE);
      EndShaderMode();
    EndTextureMode();
    test = 1;

    for (int i = cascadeAmount; i >= 0; i--) {
      radianceBufferC = radianceBufferA;
      radianceBufferA = radianceBufferB;
      radianceBufferB = radianceBufferC;

      if (rcBilinear)
        SetTextureFilter(radianceBufferC.texture, TEXTURE_FILTER_BILINEAR);
      else
        SetTextureFilter(radianceBufferC.texture, TEXTURE_FILTER_POINT);

      BeginTextureMode(radianceBufferA);
        BeginShaderMode(rcShader);
          ClearBackground(BLANK);
          SetShaderValueTexture(rcShader, GetShaderLocation(rcShader, "uDistanceField"), distFieldBuf.texture);
          SetShaderValueTexture(rcShader, GetShaderLocation(rcShader, "uSceneMap"),      tempBuf.texture);
          SetShaderValueTexture(rcShader, GetShaderLocation(rcShader, "uLastPass"),      radianceBufferC.texture);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uResolution"),          &resolution,          SHADER_UNIFORM_VEC2);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uBaseRayCount"),        &rcRayCount,          SHADER_UNIFORM_INT);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uMaxSteps"),            &maxRaySteps,         SHADER_UNIFORM_INT);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uCascadeDisplayIndex"), &cascadeDisplayIndex, SHADER_UNIFORM_INT);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uCascadeIndex"),        &i,                   SHADER_UNIFORM_INT);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uCascadeAmount"),       &cascadeAmount,       SHADER_UNIFORM_INT);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uSrgb"),                &srgbInt,             SHADER_UNIFORM_INT);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uTest"),                &test,                SHADER_UNIFORM_INT);
          SetShaderValue(rcShader, GetShaderLocation(rcShader, "uDecayRate"),  &giDecayRate, SHADER_UNIFORM_FLOAT);
          DrawRectangle(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, WHITE);
        EndShaderMode();
      EndTextureMode();
    }
  }

  resolution *= GetWindowScaleDPI();

  BeginTextureMode(lastFrameBuf);
    DrawTextureRec(radianceBufferA.texture, {0, 0.0, SCREEN_WIDTH, SCREEN_HEIGHT}, {0.0, 0.0}, WHITE);
  EndTextureMode();

  BeginShaderMode(finalShader);
    SetShaderValueTexture(finalShader, GetShaderLocation(finalShader, "uCanvas"), lastFrameBuf.texture);
    SetShaderValue(finalShader, GetShaderLocation(finalShader, "uResolution"), &resolution, SHADER_UNIFORM_VEC2);
    DrawRectangle(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, WHITE);
  EndShaderMode();

  if (!skipUIRendering) {
    DrawTextureEx(user.brush.tex,
                  Vector2{ (float)(GetMouseX() - user.brush.tex.width  / 2 * user.brushSize),
                           (float)(GetMouseY() - user.brush.tex.height / 2 * user.brushSize) },
                  0.0,
                  user.brushSize,
                  Color{ 0, 0, 0, 128} );
  }
}

// -------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void Demo::renderUI() {
  // dont draw our custom cursor if we are mousing over the UI
  ImGuiIO& io = ImGui::GetIO();
  if (!io.WantCaptureMouse) {
    ImGui::SetMouseCursor(ImGuiMouseCursor_None);
    DrawTextureEx(cursorTex,
                  Vector2{ (float)(GetMouseX() - cursorTex.width / 2 * CURSOR_SIZE),
                           (float)(GetMouseY() - cursorTex.height/ 2 * CURSOR_SIZE) },
                  0.0,
                  CURSOR_SIZE,
                  WHITE);
  }

  if (skipUIRendering) return;

  float h = 400;
  if (debug) {
    h += 180;
    if (debugShowBuffers) h += 60*4;
  }
  ImGui::SetNextWindowSize(ImVec2{250, h});
  ImGui::SetNextWindowPos(ImVec2{4, 4});

  if (!ImGui::Begin("Mode", &debugWindowData.open, debugWindowData.flags)) {
    ImGui::End();
  } else {
    if (ImGui::SmallButton("set r(a)ndom colour")) userSetRandomColor();
    Vector4 col4 = ColorNormalize(user.lightColor);
    float col[3] = { col4.x, col4.y, col4.z };
    ImGui::ColorEdit3("light color", col);
    user.lightColor = ColorFromNormalized(Vector4{col[0], col[1], col[2], 1.0});

    ImGui::SeparatorText("Scenes");

    if (ImGui::SmallButton("(c)lear")) setScene(CLEAR);
    ImGui::SameLine();
    if (ImGui::SmallButton("maze")) setScene(MAZE);
    ImGui::SameLine();
    if (ImGui::SmallButton("trees")) setScene(TREES);

    if (ImGui::SmallButton("penumbra")) setScene(PENUMBRA);
    ImGui::SameLine();
    if (ImGui::SmallButton("penumbra2")) setScene(PENUMBRA2);

    ImGui::SeparatorText("Lighting");

    ImGui::Checkbox("trad. gi", &gi);
    ImGui::SameLine();
    bool tmp = !gi;
    ImGui::Checkbox("radiance cascades", &tmp);
    gi = !tmp;

    ImGui::Checkbox("light orb circle", &orbs);
    ImGui::Checkbox("sRGB conversion", &srgb);
    ImGui::SliderFloat("##mix factor",   &giMixFactor, 0.0, 1.0, "mix factor = %f");
    ImGui::SliderFloat("##decay rate",   &giDecayRate, 0.0, 2.0, "decay rate = %f");

    if (gi) {
      ImGui::SeparatorText("traditional gi");
      ImGui::SliderInt("##rays per px",   &giRayCount, 0, 512, "ray count = %i");
      ImGui::Checkbox("noise", &giNoise);
    } else {
      ImGui::SeparatorText("radiance cascades");

      auto setParams = [this](int n){
        rcRayCount = pow(4, n);
      };

      ImGui::Text("branching factor");
      if (ImGui::SmallButton("1"))
        setParams(1);
      ImGui::SameLine();
      if (ImGui::SmallButton("2"))
        setParams(2);
      ImGui::SameLine();
      if (ImGui::SmallButton("3"))
        setParams(3);
      ImGui::SliderInt("##rays per px",   &rcRayCount, 0, 64,  "base ray count = %i");
      // ImGui::SliderInt("##rays per px",   &rcRayCount, 0, 64,  "top cascade probe amount = %i");
      ImGui::SliderInt("##cascade amount",   &cascadeAmount, 0, 32, "cascade amount = %i");
      ImGui::SliderInt("##cascade display",   &cascadeDisplayIndex, 0, cascadeAmount-1, "display cascade = %i");
      ImGui::Checkbox("bilinear interpolation", &rcBilinear);
    }

    if (debug) {
      ImGui::SeparatorText("Debug");
      ImGui::Text("%d FPS", GetFPS());

      if (ImGui::SmallButton("show buffers")) debugShowBuffers = !debugShowBuffers;
      if (debugShowBuffers) {
        if (ImGui::BeginTable("buffer_table", 2)) {
          ImGui::TableNextRow();
            ImGui::TableSetColumnIndex(0);
            rlImGuiImageSizeV(&emissionMap.tex,      {80, 60});
            ImGui::TableSetColumnIndex(1);
            rlImGuiImageSizeV(&tempBuf.texture,     {80, 60});
          ImGui::TableNextRow();
            ImGui::TableSetColumnIndex(0);
            rlImGuiImageSizeV(&sceneBuf.texture,     {80, 60});
          ImGui::TableNextRow();
            ImGui::TableSetColumnIndex(0);
            rlImGuiImageSizeV(&bufferA.texture,      {80, 60});
            ImGui::TableSetColumnIndex(1);
            rlImGuiImageSizeV(&distFieldBuf.texture, {80, 60});
          ImGui::TableNextRow();
            ImGui::TableSetColumnIndex(0);
            rlImGuiImageSizeV(&radianceBufferA.texture,      {80, 60});
            ImGui::TableSetColumnIndex(1);
            rlImGuiImageSizeV(&radianceBufferB.texture,      {80, 60});
          ImGui::TableNextRow();
            ImGui::TableSetColumnIndex(0);
            rlImGuiImageSizeV(&radianceBufferC.texture,      {80, 60});
          ImGui::EndTable();
        }
      }
      ImGui::SeparatorText("Build Info");
      std::string str = "built ";
      str += __DATE__;
      str += " at ";
      str += __TIME__;
      ImGui::Text(str.c_str());
    }
    ImGui::End();
  }
}

// -------------------------------------------------------------------------------------------------------------------------------------------------------------------------

// FIX: screenshots arent saved in a specified directory, might need to adapt raylib's screenshot function
void Demo::processKeyboardInput() {
  if (IsKeyPressed(KEY_ONE))   user.mode = DRAWING;
  if (IsKeyPressed(KEY_TWO))   user.mode = LIGHTING;

  if (IsKeyPressed(KEY_GRAVE)) debug = !debug;
  if (IsKeyPressed(KEY_F1))    skipUIRendering = !skipUIRendering;
  // if (IsKeyPressed(KEY_F2)) {
  //   printf("Taking screenshot.\n");
  //   if (!DirectoryExists("screenshots")) MakeDirectory("screenshots");
  //   std::string pwd = GetWorkingDirectory();
  //   ChangeDirectory("screenshots");
  //   TakeScreenshot("screenshot.png");
  //   ChangeDirectory(pwd.c_str());
  // }
  // if (IsKeyPressed(KEY_F12)) ToggleFullscreen();

  if (IsKeyPressed(KEY_C)) setScene(CLEAR);
  if (IsKeyPressed(KEY_R)) reload();
  if (IsKeyPressed(KEY_A)) userSetRandomColor();
  if (IsKeyPressed(KEY_H)) help = !help;
}

// -------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void Demo::processMouseInput() {
  // we do not want to be affecting the scene when we're clicking on the UI
  if (ImGui::GetIO().WantCaptureMouse) return;

  user.brushSize += GetMouseWheelMove() / 100;
  if      (user.brushSize < 0.05) user.brushSize = 0.05;
  else if (user.brushSize > 1.0)  user.brushSize = 1.0;

  if (IsMouseButtonDown(0) || IsMouseButtonDown(1)) {
    // TODO: CPU-based drawing is not very efficient
    auto draw = [this](ImageTexture canvas, Color color, Vector2 pos = GetMousePosition()) {
      ImageDraw(&canvas.img,
                this->user.brush.img,
                Rectangle{ 0, 0, (float)canvas.img.width, (float)canvas.img.height },
                Rectangle{ static_cast<float>(pos.x - this->user.brush.img.width  / 2 * this->user.brushSize),
                           static_cast<float>(pos.y - this->user.brush.img.height / 2 * this->user.brushSize),
                           static_cast<float>(this->user.brush.img.width  * this->user.brushSize),
                           static_cast<float>(this->user.brush.img.height * this->user.brushSize) },
                color);
    };

    ImageTexture& canvas = (user.mode == LIGHTING) ? emissionMap : occlusionMap;
    Color color;
    if (IsMouseButtonDown(0) && !IsKeyDown(KEY_LEFT_CONTROL)) {                                  // DRAW
      color = (user.mode == LIGHTING) ? user.lightColor : BLACK;
    } else if (IsMouseButtonDown(1) || (IsKeyDown(KEY_LEFT_CONTROL) && IsMouseButtonDown(0))) { // ERASE
      color = (user.mode == LIGHTING) ? BLACK : WHITE;
    }

    // Lerp between current mouse position and previous recorded mouse position so that we can have nice smooth lines when drawing
    // this appears to be a very expensive operation! hence why it is enclosed in this if statement
    // -20 FPS when drawing! crazy
    for (int i = 0; i < 8; i++) {
      Vector2 pos = GetSplinePointLinear(lastMousePos, GetMousePosition(), i/10.0);
      draw(canvas, color, pos);
    }
    draw(canvas, color);
    RELOAD_CANVAS();
  }

  lastMousePos = GetMousePosition();
}

// -------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void Demo::userSetRandomColor() {
  Color col = ColorFromNormalized(Vector4{(std::sin(static_cast<float>(GetTime()))   + 1) / 2,
                                          (std::cos(static_cast<float>(GetTime()))   + 1) / 2,
                                          (std::sin(static_cast<float>(GetTime())*2) + 1) / 2,
                                  1.0 });

  // at least one colour channel needs to be at maximum
  if (col.r < 255 && col.g < 255 & col.b < 255) {
    switch (GetRandomValue(0, 2)) {
      case 0:
        col.r = 255;
        break;
      case 1:
        col.g = 255;
        break;
      case 2:
        col.b = 255;
        break;
    }
  }

  user.lightColor = col;
}

void Demo::loadShader(std::string shader) {
  std::string path = "res/shaders/" + shader;
  Shader s = LoadShader(0, path.c_str());
  if (!IsShaderValid(s)) {
    std::cout << "ERR: '" << shader << "' is broken." << std::endl;
    UnloadShader(s);
    s = LoadShader("res/shaders/default.vert", "res/shaders/broken.frag");
  } else {
    std::cout << "Loaded '" << shader << "' successfully." << std::endl;
  }
  shaders[shader] = s;
}

// reload based on what mode we're in, unless we're press control
// if we're pressing control we reload shaders
//
// DRAWING: load canvas image texture
// LIGHTING: set lighting to how it is when the programme starts
void Demo::reload() {
  if (IsKeyDown(KEY_LEFT_CONTROL)) {
    std::cout << "Reloading shaders." << std::endl;
    for (auto const& [key, val] : shaders) {
      loadShader(key);
    }
  } else {
    printf("Replacing canvas.\n");
    if (user.mode == DRAWING) occlusionMap.img = GenImageColor(SCREEN_WIDTH, SCREEN_HEIGHT, WHITE);
    if (user.mode == LIGHTING) emissionMap.img = GenImageColor(SCREEN_WIDTH, SCREEN_HEIGHT, BLACK);
    RELOAD_CANVAS();
  }
}

void Demo::setScene(Scene scene) {
  switch (scene) {
    case CLEAR:
      occlusionMap.img = GenImageColor(SCREEN_WIDTH, SCREEN_HEIGHT, WHITE);
      emissionMap.img = GenImageColor(SCREEN_WIDTH, SCREEN_HEIGHT, BLACK);
      break;
    case MAZE:
      occlusionMap.img = LoadImage("res/textures/canvas/maze.png");
      // emissionMap.img = GenImageColor(SCREEN_WIDTH, SCREEN_HEIGHT, BLACK);
      break;
    case TREES:
      occlusionMap.img = LoadImage("res/textures/canvas/trees.png");
      // emissionMap.img = GenImageColor(SCREEN_WIDTH, SCREEN_HEIGHT, BLACK);
      break;
    case PENUMBRA:
      occlusionMap.img = LoadImage("res/textures/canvas/penumbra.png");
      emissionMap.img = LoadImage("res/textures/canvas/penumbra_e.png");
      break;
    case PENUMBRA2:
      occlusionMap.img = LoadImage("res/textures/canvas/penumbra2.png");
      emissionMap.img = LoadImage("res/textures/canvas/penumbra2_e.png");
      break;
  }
  RELOAD_CANVAS();
}
